package loosy

import (
	"bytes"
	"fmt"
)

// Any represents a generic json value.
type Any struct{ i interface{} }

// Field gets the field in Any{} with name, else nil.
func (a *Any) Field(name string) *Any {
	return a.Object()[name]
}

// Int returns the value of Any{} cast to an int64, else 0.
func (a *Any) Int() int64 {
	i, _ := a.Type().(int64)
	return i
}

// SetInt sets the underlying value of Any{} to given int64.
func (a *Any) SetInt(i int64) {
	a.i = i
}

// Uint returns the value of Any{} cast to an uint64, else 0.
func (a *Any) Uint() uint64 {
	u, _ := a.Type().(uint64)
	return u
}

// SetUint sets the underlying value of Any{} to given uint64.
func (a *Any) SetUint(u uint64) {
	a.i = u
}

// Float returns the value of Any{} cast to a float64, else 0.
func (a *Any) Float() float64 {
	f, _ := a.Type().(float64)
	return f
}

// SetFloat sets the underlying value of Any{} to given float64.
func (a *Any) SetFloat(f float64) {
	a.i = f
}

// String returns the value of Any{} cast to a string, else "".
func (a *Any) String() string {
	s, _ := a.Type().(string)
	return s
}

// SetString sets the underlying value of Any{} to given string.
func (a *Any) SetString(s string) {
	a.i = s
}

// Array returns the value of Any{} cast to an array, else nil.
func (a *Any) Array() []*Any {
	arr, _ := a.Type().([]*Any)
	return arr
}

// SetArray sets the underlying value of Any{} to given array.
func (a *Any) SetArray(arr []*Any) {
	a.i = arr
}

// Object returns the value of Any{} cast to an object, else nil.
func (a *Any) Object() map[string]*Any {
	obj, _ := a.Type().(map[string]*Any)
	return obj
}

// SetObject sets the underlying value of Any{} to given object.
func (a *Any) SetObject(obj map[string]*Any) {
	a.i = obj
}

// Type returns the underlying Any{} interface value.
func (a *Any) Type() interface{} {
	if a == nil {
		return nil
	}
	return a.i
}

// IterElems iterates the elements of Any{} cast as an array,
// successfuly casted as-so, or handling as a single element.
func (a *Any) IterElems(fn func(int, *Any)) {
	if fn == nil {
		panic("nil fn")
	}
	if a == nil {
		return
	}
	arr, _ := a.i.([]*Any)
	if arr == nil {
		fn(0, a)
		return
	}
	for i, elem := range arr {
		fn(i, elem)
	}
}

// IterFields iterates the fields of Any{} cast as an object.
func (a *Any) IterFields(fn func(string, *Any)) {
	if fn == nil {
		panic("nil fn")
	}
	for key, val := range a.Object() {
		fn(key, val)
	}
}

// MarshalJSON formats Any{} as json data.
func (a *Any) MarshalJSON() ([]byte, error) {
	if a == nil {
		return []byte("null"), nil
	}
	stream := config.BorrowStream(nil)
	encodeAny(stream, a)
	data := stream.Buffer()
	data = bytes.Clone(data)
	err := stream.Error
	config.ReturnStream(stream)
	return data, err
}

// UnmarshalJSON parses the given json data into Any{}.
func (a *Any) UnmarshalJSON(data []byte) error {
	if a == nil {
		return errNilReceiver
	}
	iter := config.BorrowIterator(data)
	decodeAny(iter, a)
	err := iter.Error
	config.ReturnIterator(iter)
	return err
}

// IsNull returns whether Any{} is the json "null" value.
func (a *Any) IsNull() bool {
	return a.i == nil
}

// SetNull sets the Any{} data to the json "null" value.
func (a *Any) SetNull() {
	a.i = nil
}

// IsEmpty ...
func (a *Any) IsEmpty() bool {
	return a == nil ||
		a.i == nil
}

// IsArray returns whether Any{} is an possible type.
func (a *Any) IsArray() bool {
	_, ok := a.Type().([]*Any)
	return ok
}

// IsObject returns whether Any{} is an object type.
func (a *Any) IsObject() bool {
	_, ok := a.Type().(map[string]*Any)
	return ok
}

// Format: implements fmt.Formatter for Any{}.
func (a *Any) Format(state fmt.State, verb rune) {
	if verb != 'v' {
		panic("invalid verb for Any{}")
	}
	var format string
	switch {
	case state.Flag('-'):
		format = "%-v"
	case state.Flag('+'):
		format = "%+v"
	case state.Flag('#'):
		format = "%#v"
	default:
		format = "%v"
	}
	fmt.Fprintf(state, format, a.i)
}
