package loosy

import (
	"bytes"
	"errors"
	"io"

	jsoniter "github.com/json-iterator/go"
)

var (
	// global jsoniter configuration to use.
	config = jsoniter.ConfigDefault

	// errNilReceiver is returned on attempted
	// get-type call to a nil receiving ptr type.
	errNilReceiver = errors.New("nil receiver")
)

// Marshal: see github.com/json-iterator/go.Marshal() (using the default configuration).
func Marshal(v any) ([]byte, error) {
	stream := config.BorrowStream(nil)
	stream.WriteVal(v)
	data := stream.Buffer()
	data = bytes.Clone(data)
	err := stream.Error
	config.ReturnStream(stream)
	return data, err
}

// MarshalWrite is a streamed alternative to Marshal().
func MarshalWrite(w io.Writer, v any) error {
	stream := config.BorrowStream(w)
	stream.WriteVal(v)
	err := stream.Error
	_ = stream.Flush()
	config.ReturnStream(stream)
	return err
}

// Unmarshal: see github.com/json-iterator/go.Unmarshal() (using the default configuration).
func Unmarshal(data []byte, v any) error {
	iter := config.BorrowIterator(data)
	iter.ReadVal(v)
	err := iter.Error
	config.ReturnIterator(iter)
	return err
}

// UnmarshalRead is a streamed alternative to Unmarshal().
func UnmarshalRead(r io.Reader, v any) error {
	iter := config.BorrowIterator(nil)
	iter.Reset(r)
	iter.ReadVal(v)
	err := iter.Error
	config.ReturnIterator(iter)
	return err
}

// JSON represents unparsed json data,
// the helper methods allow you to lazily
// parse the data *as* you need it.
type JSON struct{ B []byte }

// Any unmarshals the data in JSON{} as the Any{} type.
func (j *JSON) Any() (a *Any, err error) {
	if j == nil {
		return nil, errNilReceiver
	}
	a = new(Any) // alloc new Any
	iter := config.BorrowIterator(j.B)
	decodeAny(iter, a)
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetAny sets the data in JSON{} to marshaled Any{} bytes.
func (j *JSON) SetAny(a *Any) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	encodeAny(stream, a)
	// err = stream.Error
	config.ReturnStream(stream)
}

// Bool unmarshals the data in JSON{} as bool.
func (j *JSON) Bool() (b bool, err error) {
	if j == nil {
		return false, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	b = iter.ReadBool()
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetBool sets the data in JSON{} to marshaled bool bytes.
func (j *JSON) SetBool(b bool) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteBool(b)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// Int unmarshals the data in JSON{} as int64.
func (j *JSON) Int() (i int64, err error) {
	if j == nil {
		return 0, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	i = iter.ReadInt64()
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetInt sets the data in JSON{} to marshaled int64 bytes.
func (j *JSON) SetInt(i int64) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteInt64(i)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// Uint unmarshals the data in JSON{} as uint64.
func (j *JSON) Uint() (u uint64, err error) {
	if j == nil {
		return 0, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	u = iter.ReadUint64()
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetUint sets the data in JSON{} to marshaled uint64 bytes.
func (j *JSON) SetUint(u uint64) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteUint64(u)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// Float unmarshals the data in JSON{} as float64.
func (j *JSON) Float() (f float64, err error) {
	if j == nil {
		return 0, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	f = iter.ReadFloat64()
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetFloat sets the data in JSON{} to marshaled float64 bytes.
func (j *JSON) SetFloat(f float64) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteFloat64(f)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// SetFloatLossy sets the data in JSON{} to marshaled float64 bytes (much faster, though with only 6 digits precision).
func (j *JSON) SetFloatLossy(f float64) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteFloat64Lossy(f)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// String unmarshals the data in JSON{} as string.
func (j *JSON) String() (s string, err error) {
	if j == nil {
		return "", errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	s = iter.ReadString()
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetString sets the data in JSON{} to marshaled string bytes.
func (j *JSON) SetString(s string) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteString(s)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// SetString sets the data in JSON{} to marshaled string bytes (with HTML special chars escaped).
func (j *JSON) SetStringHTMLEscaped(s string) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteStringWithHTMLEscaped(s)
	// err = stream.Error
	stream.Flush()
	config.ReturnStream(stream)
}

// Array unmarshals the data in JSON{} as []*JSON{}.
func (j *JSON) Array() (arr []*JSON, err error) {
	if j == nil {
		return nil, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	for iter.ReadArray() {
		elem := new(JSON)
		decodeJSON(iter, elem)
		arr = append(arr, elem)
	}
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetArray sets the data in JSON{} to marshaled []*JSON{} bytes.
func (j *JSON) SetArray(arr []*JSON) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)

	if l := len(arr); l == 0 {
		stream.WriteEmptyArray()
	} else {
		stream.WriteArrayStart()
		for i := 0; i < len(arr)-1; i++ {
			encodeJSON(stream, arr[i])
			stream.WriteMore()
		}
		encodeJSON(stream, arr[l-1])
		stream.WriteArrayEnd()
	}

	// err = stream.Error
	_ = stream.Flush()
	config.ReturnStream(stream)
}

// Object unmarshals the data in JSON{} as object.
func (j *JSON) Object() (obj map[string]*JSON, err error) {
	if j == nil {
		return nil, errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	decodeJSONObject(iter, &obj)
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetObject sets the data in JSON{} to marshaled object bytes.
func (j *JSON) SetObject(obj map[string]*JSON) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	encodeJSONObject(stream, obj)
	// err = stream.Error
	_ = stream.Flush()
	config.ReturnStream(stream)
}

// Value unmarshals the data in JSON{} into value 'v'.
func (j *JSON) Value(v any) (err error) {
	if j == nil {
		return errNilReceiver
	}
	iter := config.BorrowIterator(j.B)
	iter.ReadVal(v)
	err = iter.Error
	config.ReturnIterator(iter)
	return
}

// SetValue sets the data in JSON{} to marshaled value 'v' bytes.
func (j *JSON) SetValue(v any) (err error) {
	j.B = j.B[:0] // reset before write.
	stream := config.BorrowStream(j)
	stream.WriteVal(v)
	err = stream.Error
	_ = stream.Flush()
	config.ReturnStream(stream)
	return
}

// MarshalJSON directly returns the data stored in JSON{}, or the json "null" value if nil.
func (j *JSON) MarshalJSON() ([]byte, error) {
	if j == nil {
		return []byte("null"), nil
	}
	return j.B, nil
}

// UnmarshalJSON copies the data into JSON{}, delaying actual unmarshaling.
func (j *JSON) UnmarshalJSON(data []byte) error {
	if j == nil {
		return errNilReceiver
	}
	j.B = append(j.B[:0], data...)
	return nil
}

// IsNull returns whether JSON{} data is the json "null" value.
func (j *JSON) IsNull() bool {
	return string(j.B) == "null"
}

// SetNull sets the JSON{} data to the json "null" value.
func (j *JSON) SetNull() {
	j.B = []byte("null")
}

// IsEmpty ...
func (j *JSON) IsEmpty() bool {
	return j == nil ||
		len(j.B) == 0 ||
		j.IsNull()
}

// IsArray returns whether JSON{} is a possible array (i.e. has surrounding square braces).
func (j *JSON) IsArray() bool {
	return len(j.B) >= 2 && j.B[0] == '[' && j.B[len(j.B)-1] == ']'
}

// IsObject returns whether JSON{} is a possible object (i.e. has surrounding curl braces).
func (j *JSON) IsObject() bool {
	return len(j.B) >= 2 && j.B[0] == '{' && j.B[len(j.B)-1] == '}'
}

// Write: implements io.Writer for JSON{}.
func (j *JSON) Write(p []byte) (n int, err error) {
	j.B = append(j.B, p...)
	return len(p), nil
}
